@@ -2,10 +2,28 @@ import json
 import logging
 import os
 import weakref
-from typing import Any, Callable, Dict, Generic, NamedTuple, Optional, Union, overload
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generic,
+    NamedTuple,
+    Optional,
+    Union,
+    cast,
+    overload,
+)
 
 from aiortc.mediastreams import MediaStreamTrack
 
+from streamlit_webrtc.models import (
+    AudioFrameCallback,
+    MediaEndedCallback,
+    QueuedAudioFramesCallback,
+    QueuedVideoFramesCallback,
+    VideoFrameCallback,
+)
+
 try:
     from typing import TypedDict
 except ImportError:
@@ -110,7 +128,12 @@ class WebRtcStreamerContext(Generic[VideoProcessorT, AudioProcessorT]):
         to `webrtc_streamer()`.
         """
         worker = self._get_worker()
-        return worker.video_processor if worker else None
+
+        # worker.video_processor is Union[VideoProcessorT, MediaCallbackContainer],
+        # but the callback processor is usually not used through this property
+        # as class-less callback API is used in that case,
+        # so we can ignore that type here by casting the type into VideoProcessorT only.
+        return cast(VideoProcessorT, worker.video_processor) if worker else None
 
     @property
     def audio_processor(self) -> Optional[AudioProcessorT]:
@@ -120,7 +143,12 @@ class WebRtcStreamerContext(Generic[VideoProcessorT, AudioProcessorT]):
         to `webrtc_streamer()`.
         """
         worker = self._get_worker()
-        return worker.audio_processor if worker else None
+
+        # worker.audio_processor is Union[AudioProcessorT, MediaCallbackContainer],
+        # but the callback processor is usually not used through this property
+        # as class-less callback API is used in that case,
+        # so we can ignore that type here by casting the type into AudioProcessorT only.
+        return cast(AudioProcessorT, worker.audio_processor) if worker else None
 
     @property
     def video_transformer(self) -> Optional[VideoProcessorT]:
@@ -132,7 +160,7 @@ class WebRtcStreamerContext(Generic[VideoProcessorT, AudioProcessorT]):
         .. deprecated:: 0.20.0
         """
         worker = self._get_worker()
-        return worker.video_processor if worker else None
+        return cast(VideoProcessorT, worker.video_processor) if worker else None
 
     @property
     def video_receiver(self) -> Optional[VideoReceiver]:
@@ -198,6 +226,12 @@ def webrtc_streamer(
     player_factory: Optional[MediaPlayerFactory] = None,
     in_recorder_factory: Optional[MediaRecorderFactory] = None,
     out_recorder_factory: Optional[MediaRecorderFactory] = None,
+    video_frame_callback: Optional[VideoFrameCallback] = None,
+    audio_frame_callback: Optional[AudioFrameCallback] = None,
+    queued_video_frames_callback: Optional[QueuedVideoFramesCallback] = None,
+    queued_audio_frames_callback: Optional[QueuedAudioFramesCallback] = None,
+    on_video_ended: Optional[MediaEndedCallback] = None,
+    on_audio_ended: Optional[MediaEndedCallback] = None,
     video_processor_factory: None = None,
     audio_processor_factory: None = None,
     async_processing: bool = True,
@@ -233,6 +267,12 @@ def webrtc_streamer(
     player_factory: Optional[MediaPlayerFactory] = None,
     in_recorder_factory: Optional[MediaRecorderFactory] = None,
     out_recorder_factory: Optional[MediaRecorderFactory] = None,
+    video_frame_callback: Optional[VideoFrameCallback] = None,
+    audio_frame_callback: Optional[AudioFrameCallback] = None,
+    queued_video_frames_callback: Optional[QueuedVideoFramesCallback] = None,
+    queued_audio_frames_callback: Optional[QueuedAudioFramesCallback] = None,
+    on_video_ended: Optional[MediaEndedCallback] = None,
+    on_audio_ended: Optional[MediaEndedCallback] = None,
     video_processor_factory: Optional[VideoProcessorFactory[VideoProcessorT]] = None,
     audio_processor_factory: None = None,
     async_processing: bool = True,
@@ -264,6 +304,12 @@ def webrtc_streamer(
     player_factory: Optional[MediaPlayerFactory] = None,
     in_recorder_factory: Optional[MediaRecorderFactory] = None,
     out_recorder_factory: Optional[MediaRecorderFactory] = None,
+    video_frame_callback: Optional[VideoFrameCallback] = None,
+    audio_frame_callback: Optional[AudioFrameCallback] = None,
+    queued_video_frames_callback: Optional[QueuedVideoFramesCallback] = None,
+    queued_audio_frames_callback: Optional[QueuedAudioFramesCallback] = None,
+    on_video_ended: Optional[MediaEndedCallback] = None,
+    on_audio_ended: Optional[MediaEndedCallback] = None,
     video_processor_factory: None = None,
     audio_processor_factory: Optional[AudioProcessorFactory[AudioProcessorT]] = None,
     async_processing: bool = True,
@@ -295,6 +341,12 @@ def webrtc_streamer(
     player_factory: Optional[MediaPlayerFactory] = None,
     in_recorder_factory: Optional[MediaRecorderFactory] = None,
     out_recorder_factory: Optional[MediaRecorderFactory] = None,
+    video_frame_callback: Optional[VideoFrameCallback] = None,
+    audio_frame_callback: Optional[AudioFrameCallback] = None,
+    queued_video_frames_callback: Optional[QueuedVideoFramesCallback] = None,
+    queued_audio_frames_callback: Optional[QueuedAudioFramesCallback] = None,
+    on_video_ended: Optional[MediaEndedCallback] = None,
+    on_audio_ended: Optional[MediaEndedCallback] = None,
     video_processor_factory: Optional[VideoProcessorFactory[VideoProcessorT]] = None,
     audio_processor_factory: Optional[AudioProcessorFactory[AudioProcessorT]] = None,
     async_processing: bool = True,
@@ -325,6 +377,12 @@ def webrtc_streamer(
     player_factory: Optional[MediaPlayerFactory] = None,
     in_recorder_factory: Optional[MediaRecorderFactory] = None,
     out_recorder_factory: Optional[MediaRecorderFactory] = None,
+    video_frame_callback: Optional[VideoFrameCallback] = None,
+    audio_frame_callback: Optional[AudioFrameCallback] = None,
+    queued_video_frames_callback: Optional[QueuedVideoFramesCallback] = None,
+    queued_audio_frames_callback: Optional[QueuedAudioFramesCallback] = None,
+    on_video_ended: Optional[MediaEndedCallback] = None,
+    on_audio_ended: Optional[MediaEndedCallback] = None,
     video_processor_factory=None,
     audio_processor_factory=None,
     async_processing: bool = True,
@@ -487,6 +545,20 @@ def webrtc_streamer(
         # Rerun to unset the SDP answer from the frontend args
         st.experimental_rerun()
 
+    if webrtc_worker:
+        if video_frame_callback or queued_video_frames_callback or on_video_ended:
+            webrtc_worker.update_video_callbacks(
+                frame_callback=video_frame_callback,
+                queued_frames_callback=queued_video_frames_callback,
+                on_ended=on_video_ended,
+            )
+        if audio_frame_callback or queued_audio_frames_callback or on_audio_ended:
+            webrtc_worker.update_audio_callbacks(
+                frame_callback=audio_frame_callback,
+                queued_frames_callback=queued_audio_frames_callback,
+                on_ended=on_audio_ended,
+            )
+
     if not webrtc_worker and sdp_offer:
         LOGGER.debug(
             "No worker exists though the offer SDP is set. "
@@ -498,6 +570,12 @@ def webrtc_streamer(
             player_factory=player_factory,
             in_recorder_factory=in_recorder_factory,
             out_recorder_factory=out_recorder_factory,
+            video_frame_callback=video_frame_callback,
+            audio_frame_callback=audio_frame_callback,
+            queued_video_frames_callback=queued_video_frames_callback,
+            queued_audio_frames_callback=queued_audio_frames_callback,
+            on_video_ended=on_video_ended,
+            on_audio_ended=on_audio_ended,
             video_processor_factory=video_processor_factory,
             audio_processor_factory=audio_processor_factory,
             async_processing=async_processing,
